@@ -1,9 +1,9 @@
 #! /bin/sh
 # check script for Plzip - Massively parallel implementation of lzip
-# Copyright (C) 2009-2019 Antonio Diaz Diaz.
+# Copyright (C) 2009-2021 Antonio Diaz Diaz.
 #
 # This script is free software: you have unlimited permission
-# to copy, distribute and modify it.
+# to copy, distribute, and modify it.
 
 LC_ALL=C
 export LC_ALL
@@ -30,6 +30,7 @@ cd "${objdir}"/tmp || framework_failure
 
 cat "${testdir}"/test.txt > in || framework_failure
 in_lz="${testdir}"/test.txt.lz
+in_em="${testdir}"/test_em.txt.lz
 fail=0
 lwarn8=0
 lwarn10=0
@@ -41,6 +42,7 @@ lzlib_1_10() { [ ${lwarn10} = 0 ] &&
 	printf "\nwarning: header HD=3 detection requires lzlib 1.10 or newer"
 	lwarn10=1 ; }
 
+"${LZIP}" --check-lib				# just print warning
 printf "testing plzip-%s..." "$2"
 
 "${LZIP}" -fkqm4 in
@@ -66,6 +68,14 @@ done
 [ $? = 2 ] || test_failed $LINENO
 "${LZIP}" -dq -o in < "${in_lz}"
 [ $? = 1 ] || test_failed $LINENO
+"${LZIP}" -dq -o in "${in_lz}"
+[ $? = 1 ] || test_failed $LINENO
+"${LZIP}" -dq -o out nx_file.lz
+[ $? = 1 ] || test_failed $LINENO
+[ ! -e out ] || test_failed $LINENO
+"${LZIP}" -q -o out.lz nx_file
+[ $? = 1 ] || test_failed $LINENO
+[ ! -e out.lz ] || test_failed $LINENO
 # these are for code coverage
 "${LZIP}" -lt "${in_lz}" 2> /dev/null
 [ $? = 1 ] || test_failed $LINENO
@@ -73,7 +83,9 @@ done
 [ $? = 1 ] || test_failed $LINENO
 "${LZIP}" -cdt "${in_lz}" > out 2> /dev/null
 [ $? = 1 ] || test_failed $LINENO
-"${LZIP}" -t -- nx_file 2> /dev/null
+"${LZIP}" -t -- nx_file.lz 2> /dev/null
+[ $? = 1 ] || test_failed $LINENO
+"${LZIP}" -t "" < /dev/null 2> /dev/null
 [ $? = 1 ] || test_failed $LINENO
 "${LZIP}" --help > /dev/null || test_failed $LINENO
 "${LZIP}" -n1 -V > /dev/null || test_failed $LINENO
@@ -97,12 +109,26 @@ printf "LZIP\001+.............................." | "${LZIP}" -t 2> /dev/null
 
 printf "\ntesting decompression..."
 
-"${LZIP}" -lq "${in_lz}" || test_failed $LINENO
-"${LZIP}" -t "${in_lz}" || test_failed $LINENO
-"${LZIP}" -cd "${in_lz}" > copy || test_failed $LINENO
-cmp in copy || test_failed $LINENO
+for i in "${in_lz}" "${in_em}" ; do
+	"${LZIP}" -lq "$i" || test_failed $LINENO "$i"
+	"${LZIP}" -t "$i" || test_failed $LINENO "$i"
+	"${LZIP}" -d "$i" -o copy || test_failed $LINENO "$i"
+	cmp in copy || test_failed $LINENO "$i"
+	"${LZIP}" -cd "$i" > copy || test_failed $LINENO "$i"
+	cmp in copy || test_failed $LINENO "$i"
+	"${LZIP}" -d "$i" -o - > copy || test_failed $LINENO "$i"
+	cmp in copy || test_failed $LINENO "$i"
+	"${LZIP}" -d < "$i" > copy || test_failed $LINENO "$i"
+	cmp in copy || test_failed $LINENO "$i"
+	rm -f copy || framework_failure
+done
+
+lines=$("${LZIP}" -tvv "${in_em}" 2>&1 | wc -l) || test_failed $LINENO
+[ "${lines}" -eq 1 ] || test_failed $LINENO "${lines}"
+
+lines=$("${LZIP}" -lvv "${in_em}" | wc -l) || test_failed $LINENO
+[ "${lines}" -eq 11 ] || test_failed $LINENO "${lines}"
 
-rm -f copy || framework_failure
 cat "${in_lz}" > copy.lz || framework_failure
 "${LZIP}" -dk copy.lz || test_failed $LINENO
 cmp in copy || test_failed $LINENO
@@ -113,19 +139,19 @@ printf "to be overwritten" > copy || framework_failure
 [ ! -e copy.lz ] || test_failed $LINENO
 cmp in copy || test_failed $LINENO
 
-rm -f copy || framework_failure
-cat "${in_lz}" > copy.lz || framework_failure
-"${LZIP}" -d -S100k copy.lz || test_failed $LINENO	# ignore -S
-[ ! -e copy.lz ] || test_failed $LINENO
-cmp in copy || test_failed $LINENO
-
 printf "to be overwritten" > copy || framework_failure
 "${LZIP}" -df -o copy < "${in_lz}" || test_failed $LINENO
 cmp in copy || test_failed $LINENO
+rm -f out copy || framework_failure
+"${LZIP}" -d -o ./- "${in_lz}" || test_failed $LINENO
+cmp in ./- || test_failed $LINENO
+rm -f ./- || framework_failure
+"${LZIP}" -d -o ./- < "${in_lz}" || test_failed $LINENO
+cmp in ./- || test_failed $LINENO
+rm -f ./- || framework_failure
 
-rm -f copy || framework_failure
-"${LZIP}" < in > anyothername || test_failed $LINENO
-"${LZIP}" -dv --output copy - anyothername - < "${in_lz}" 2> /dev/null ||
+cat "${in_lz}" > anyothername || framework_failure
+"${LZIP}" -dv - anyothername - < "${in_lz}" > copy 2> /dev/null ||
 	test_failed $LINENO
 cmp in copy || test_failed $LINENO
 cmp in anyothername.out || test_failed $LINENO
@@ -166,21 +192,19 @@ done
 cmp in copy || test_failed $LINENO
 
 cat in in > in2 || framework_failure
-cat "${in_lz}" "${in_lz}" > in2.lz || framework_failure
-"${LZIP}" -lq in2.lz || test_failed $LINENO
-"${LZIP}" -t in2.lz || test_failed $LINENO
-"${LZIP}" -cd in2.lz > copy2 || test_failed $LINENO
+"${LZIP}" -lq "${in_lz}" "${in_lz}" || test_failed $LINENO
+"${LZIP}" -t "${in_lz}" "${in_lz}" || test_failed $LINENO
+"${LZIP}" -cd "${in_lz}" "${in_lz}" -o out > copy2 || test_failed $LINENO
+[ ! -e out ] || test_failed $LINENO			# override -o
 cmp in2 copy2 || test_failed $LINENO
-
-"${LZIP}" --output=copy2.lz < in2 || test_failed $LINENO
-"${LZIP}" -lq copy2.lz || test_failed $LINENO
-"${LZIP}" -t copy2.lz || test_failed $LINENO
-"${LZIP}" -cd copy2.lz > copy2 || test_failed $LINENO
+rm -f copy2 || framework_failure
+"${LZIP}" -d "${in_lz}" "${in_lz}" -o copy2 || test_failed $LINENO
 cmp in2 copy2 || test_failed $LINENO
+rm -f copy2 || framework_failure
 
+cat "${in_lz}" "${in_lz}" > copy2.lz || framework_failure
 printf "\ngarbage" >> copy2.lz || framework_failure
 "${LZIP}" -tvvvv copy2.lz 2> /dev/null || test_failed $LINENO
-rm -f copy2 || framework_failure
 "${LZIP}" -alq copy2.lz
 [ $? = 2 ] || test_failed $LINENO
 "${LZIP}" -atq copy2.lz
@@ -202,37 +226,46 @@ printf "\ntesting   compression..."
 
 "${LZIP}" -cf "${in_lz}" > out 2> /dev/null	# /dev/null is a tty on OS/2
 [ $? = 1 ] || test_failed $LINENO
-"${LZIP}" -cFvvm36 "${in_lz}" > out 2> /dev/null || test_failed $LINENO
+"${LZIP}" -Fvvm36 -o - "${in_lz}" > out 2> /dev/null || test_failed $LINENO
 "${LZIP}" -cd out | "${LZIP}" -d > copy || test_failed $LINENO
 cmp in copy || test_failed $LINENO
 
+"${LZIP}" -0 -o ./- in || test_failed $LINENO
+"${LZIP}" -cd ./- | cmp in - || test_failed $LINENO
+rm -f ./- || framework_failure
+"${LZIP}" -0 -o ./- < in || test_failed $LINENO		# add .lz
+[ ! -e ./- ] || test_failed $LINENO
+"${LZIP}" -cd -- -.lz | cmp in - || test_failed $LINENO
+rm -f ./-.lz || framework_failure
+
 for i in s4Ki 0 1 2 3 4 5 6 7 8 9 ; do
 	"${LZIP}" -k -$i in || test_failed $LINENO $i
 	mv -f in.lz copy.lz || test_failed $LINENO $i
 	printf "garbage" >> copy.lz || framework_failure
 	"${LZIP}" -df copy.lz || test_failed $LINENO $i
 	cmp in copy || test_failed $LINENO $i
-done
 
-for i in s4Ki 0 1 2 3 4 5 6 7 8 9 ; do
-	"${LZIP}" -c -$i in > out || test_failed $LINENO $i
+	"${LZIP}" -$i in -c > out || test_failed $LINENO $i
+	"${LZIP}" -$i in -o o_out || test_failed $LINENO $i	# don't add .lz
+	[ ! -e o_out.lz ] || test_failed $LINENO
+	cmp out o_out || test_failed $LINENO $i
+	rm -f o_out || framework_failure
 	printf "g" >> out || framework_failure
 	"${LZIP}" -cd out > copy || test_failed $LINENO $i
 	cmp in copy || test_failed $LINENO $i
-done
 
-for i in s4Ki 0 1 2 3 4 5 6 7 8 9 ; do
 	"${LZIP}" -$i < in > out || test_failed $LINENO $i
 	"${LZIP}" -d < out > copy || test_failed $LINENO $i
 	cmp in copy || test_failed $LINENO $i
-done
 
-for i in s4Ki 0 1 2 3 4 5 6 7 8 9 ; do
-	"${LZIP}" -f -$i -o out < in || test_failed $LINENO $i
+	rm -f out || framework_failure
+	printf "to be overwritten" > out.lz || framework_failure
+	"${LZIP}" -f -$i -o out < in || test_failed $LINENO $i	# add .lz
+	[ ! -e out ] || test_failed $LINENO
 	"${LZIP}" -df -o copy < out.lz || test_failed $LINENO $i
 	cmp in copy || test_failed $LINENO $i
 done
-rm -f out.lz || framework_failure
+rm -f out out.lz || framework_failure
 
 cat in in in in > in4 || framework_failure
 for i in 1 2 3 4 5 6 7 8 9 10 11 12 13 14 15 16 ; do
@@ -317,6 +350,13 @@ else
 fi
 rm -f int.lz || framework_failure
 
+for i in fox_v2.lz fox_s11.lz fox_de20.lz \
+         fox_bcrc.lz fox_crc0.lz fox_das46.lz fox_mes81.lz ; do
+	"${LZIP}" -tq "${testdir}"/$i
+	[ $? = 2 ] || test_failed $LINENO $i
+done
+
+cat "${in_lz}" "${in_lz}" > in2.lz || framework_failure
 cat "${in_lz}" "${in_lz}" "${in_lz}" > in3.lz || framework_failure
 if dd if=in3.lz of=trunc.lz bs=14752 count=1 2> /dev/null &&
    [ -e trunc.lz ] && cmp in2.lz trunc.lz > /dev/null 2>&1 ; then
@@ -343,14 +383,22 @@ printf "g" >> ingin.lz || framework_failure
 cat "${in_lz}" >> ingin.lz || framework_failure
 "${LZIP}" -lq ingin.lz
 [ $? = 2 ] || test_failed $LINENO
+"${LZIP}" -atq ingin.lz
+[ $? = 2 ] || test_failed $LINENO
+"${LZIP}" -atq < ingin.lz
+[ $? = 2 ] || test_failed $LINENO
+"${LZIP}" -acdq ingin.lz > out
+[ $? = 2 ] || test_failed $LINENO
+"${LZIP}" -adq < ingin.lz > out
+[ $? = 2 ] || test_failed $LINENO
 "${LZIP}" -tq ingin.lz
 [ $? = 2 ] || test_failed $LINENO
-"${LZIP}" -cdq ingin.lz > out
-[ $? = 2 ] || test_failed $LINENO
 "${LZIP}" -t < ingin.lz || test_failed $LINENO
+"${LZIP}" -cdq ingin.lz > copy
+[ $? = 2 ] || test_failed $LINENO
 "${LZIP}" -d < ingin.lz > copy || test_failed $LINENO
 cmp in copy || test_failed $LINENO
-rm -f copy ingin.lz || framework_failure
+rm -f copy ingin.lz out || framework_failure
 
 echo
 if [ ${fail} = 0 ] ; then
